from typing import Optional, NamedTuple
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPrivateKey,
    generate_private_key,
    RSAPrivateKeyWithSerialization,
    RSAPublicKey,
)
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    load_pem_private_key,
    load_pem_public_key,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
)


PUBLIC_EXPONENT = 65537
KEY_SIZE = 4096
PADDING = 11
MAX_BITS = int(KEY_SIZE / 8) - PADDING
ISO_ENCODING = "ISO-8859-1"
BYTE_ORDER = "big"


class RSAKeyPair(NamedTuple):
    """Key pair of RSA pkcs1 in bytes"""

    private_key: str
    public_key: str


def rsa_keypair() -> RSAKeyPair:
    """
    Create RSA keypair

    :return: RSA key pair
    """

    private_key: RSAPrivateKeyWithSerialization = generate_private_key(
        public_exponent=PUBLIC_EXPONENT, key_size=KEY_SIZE, backend=default_backend()
    )
    private_key_bytes = private_key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=NoEncryption(),
    )
    private_key_string = str(private_key_bytes, ISO_ENCODING)

    public_key: RSAPublicKey = private_key.public_key()
    public_key_bytes = public_key.public_bytes(
        encoding=Encoding.PEM, format=PublicFormat.SubjectPublicKeyInfo
    )
    public_key_string = str(public_key_bytes, ISO_ENCODING)

    return RSAKeyPair(private_key_string, public_key_string)


def rsa_encrypt(message: str, public_key: str) -> Optional[str]:
    """
    Encrypt with RSA public key

    :param message: Message
    :param key_pair: RSA public key
    :return: Encrypted message
    """
    data = bytes(public_key, ISO_ENCODING)
    rsa_public_key = load_pem_public_key(data, backend=default_backend())
    if not isinstance(rsa_public_key, RSAPublicKey):
        return None
    plaintext = bytes.fromhex(message)
    if len(plaintext) > MAX_BITS:
        return None
    ciphertext = rsa_public_key.encrypt(plaintext, PKCS1v15())
    return str(ciphertext, ISO_ENCODING)


def rsa_decrypt(encrypted_message: str, private_key: str) -> Optional[str]:
    """
    Decrypt with RSA private key

    :param encrypted_message: Encrypted message
    :param private_key: RSA private key
    :return: Message
    """

    data = bytes(private_key, ISO_ENCODING)
    rsa_private_key = load_pem_private_key(
        data, password=None, backend=default_backend()
    )
    if not isinstance(rsa_private_key, RSAPrivateKey):
        return None
    ciphertext = bytes(encrypted_message, ISO_ENCODING)
    try:
        plaintext = rsa_private_key.decrypt(ciphertext, PKCS1v15())
    except ValueError:
        return None
    hex_str = plaintext.hex()
    return hex_str



# pylint: disable=no-name-in-module
# Support for basic modular math in ElectionGuard. This code's primary purpose is to be "correct",
# in the sense that performance may be less than hand-optimized C code, and no guarantees are
# made about timing or other side-channels.

from typing import Any, Final, NamedTuple, Optional, Union
from base64 import b16decode
from secrets import randbelow
from gmpy2 import mpz, powmod, invert, to_binary, from_binary

# Constants used by ElectionGuard
Q: Final[int] = pow(2, 256) - 189
P: Final[
    int
] = 1044388881413152506691752710716624382579964249047383780384233483283953907971553643537729993126875883902173634017777416360502926082946377942955704498542097614841825246773580689398386320439747911160897731551074903967243883427132918813748016269754522343505285898816777211761912392772914485521155521641049273446207578961939840619466145806859275053476560973295158703823395710210329314709715239251736552384080845836048778667318931418338422443891025911884723433084701207771901944593286624979917391350564662632723703007964229849154756196890615252286533089643184902706926081744149289517418249153634178342075381874131646013444796894582106870531535803666254579602632453103741452569793905551901541856173251385047414840392753585581909950158046256810542678368121278509960520957624737942914600310646609792665012858397381435755902851312071248102599442308951327039250818892493767423329663783709190716162023529669217300939783171415808233146823000766917789286154006042281423733706462905243774854543127239500245873582012663666430583862778167369547603016344242729592244544608279405999759391099775667746401633668308698186721172238255007962658564443858927634850415775348839052026675785694826386930175303143450046575460843879941791946313299322976993405829119  # pylint: disable=line-too-long
R: Final[int] = ((P - 1) * pow(Q, -1, P)) % P
G: Final[
    int
] = 14245109091294741386751154342323521003543059865261911603340669522218159898070093327838595045175067897363301047764229640327930333001123401070596314469603183633790452807428416775717923182949583875381833912370889874572112086966300498607364501764494811956017881198827400327403252039184448888877644781610594801053753235453382508543906993571248387749420874609737451803650021788641249940534081464232937193671929586747339353451021712752406225276255010281004857233043241332527821911604413582442915993833774890228705495787357234006932755876972632840760599399514028393542345035433135159511099877773857622699742816228063106927776147867040336649025152771036361273329385354927395836330206311072577683892664475070720408447257635606891920123791602538518516524873664205034698194561673019535564273204744076336022130453963648114321050173994259620611015189498335966173440411967562175734606706258335095991140827763942280037063180207172918769921712003400007923888084296685269233298371143630883011213745082207405479978418089917768242592557172834921185990876960527013386693909961093302289646193295725135238595082039133488721800071459503353417574248679728577942863659802016004283193163470835709405666994892499382890912238098413819320185166580019604608311466  # pylint: disable=line-too-long
Q_MINUS_ONE: Final[int] = Q - 1


class ElementModQ(NamedTuple):
    """An element of the smaller `mod q` space, i.e., in [0, Q), where Q is a 256-bit prime."""

    elem: mpz

    def to_bytes(self) -> bytes:
        """
        Converts from the element to the representation of bytes by first going through hex.
        This is preferable to directly accessing `elem`, whose representation might change.
        """
        return b16decode(self.to_hex())

    def to_hex(self) -> str:
        """
        Converts from the element to the hex representation of bytes. This is preferable to directly
        accessing `elem`, whose representation might change.
        """
        h = format(self.elem, "02X")
        if len(h) % 2:
            h = "0" + h
        return h

    def to_int(self) -> int:
        """
        Converts from the element to a regular integer. This is preferable to directly
        accessing `elem`, whose representation might change.
        """
        return self.elem

    def is_in_bounds(self) -> bool:
        """
        Validates that the element is actually within the bounds of [0,Q).
        Returns true if all is good, false if something's wrong.
        """
        return 0 <= self.elem < Q

    def is_in_bounds_no_zero(self) -> bool:
        """
        Validates that the element is actually within the bounds of [1,Q).
        Returns true if all is good, false if something's wrong.
        """
        return 0 < self.elem < Q

    # overload != (not equal to) operator
    def __ne__(self, other: Any) -> bool:
        return (isinstance(other, (ElementModP, ElementModQ))) and not eq_elems(
            self, other
        )

    # overload == (equal to) operator
    def __eq__(self, other: Any) -> bool:
        return (isinstance(other, (ElementModP, ElementModQ))) and eq_elems(self, other)

    def __str__(self) -> str:
        return self.elem.digits()


class ElementModP(NamedTuple):
    """An element of the larger `mod p` space, i.e., in [0, P), where P is a 4096-bit prime."""

    elem: mpz

    def to_hex(self) -> str:
        """
        Converts from the element to the hex representation of bytes. This is preferable to directly
        accessing `elem`, whose representation might change.
        """
        h = format(self.elem, "02X")
        if len(h) % 2:
            h = "0" + h
        return h

    def to_int(self) -> int:
        """
        Converts from the element to a regular integer. This is preferable to directly
        accessing `elem`, whose representation might change.
        """
        return self.elem

    def is_in_bounds(self) -> bool:
        """
        Validates that the element is actually within the bounds of [0,P).
        Returns true if all is good, false if something's wrong.
        """
        return 0 <= self.elem < P

    def is_in_bounds_no_zero(self) -> bool:
        """
        Validates that the element is actually within the bounds of [1,P).
        Returns true if all is good, false if something's wrong.
        """
        return 0 < self.elem < P

    def is_valid_residue(self) -> bool:
        """
        Validates that this element is in Z^r_p.
        Returns true if all is good, false if something's wrong.
        """
        residue = pow_p(self, ElementModQ(mpz(Q))) == ONE_MOD_P
        return self.is_in_bounds() and residue

    # overload != (not equal to) operator
    def __ne__(self, other: Any) -> bool:
        return (isinstance(other, (ElementModP, ElementModQ))) and not eq_elems(
            self, other
        )

    # overload == (equal to) operator
    def __eq__(self, other: Any) -> bool:
        return (isinstance(other, (ElementModP, ElementModQ))) and eq_elems(self, other)

    def __str__(self) -> str:
        return self.elem.digits()


# Common constants
ZERO_MOD_Q: Final[ElementModQ] = ElementModQ(mpz(0))
ONE_MOD_Q: Final[ElementModQ] = ElementModQ(mpz(1))
TWO_MOD_Q: Final[ElementModQ] = ElementModQ(mpz(2))

ZERO_MOD_P: Final[ElementModP] = ElementModP(mpz(0))
ONE_MOD_P: Final[ElementModP] = ElementModP(mpz(1))
TWO_MOD_P: Final[ElementModP] = ElementModP(mpz(2))

ElementModPOrQ = Union[ElementModP, ElementModQ]
ElementModPOrQorInt = Union[ElementModP, ElementModQ, int]
ElementModQorInt = Union[ElementModQ, int]
ElementModPorInt = Union[ElementModP, int]


def hex_to_q(input: str) -> Optional[ElementModQ]:
    """
    Given a hex string representing bytes, returns an ElementModQ.
    Returns `None` if the number is out of the allowed
    [0,Q) range.
    """
    i = int(input, 16)
    if 0 <= i < Q:
        return ElementModQ(mpz(i))
    return None


def int_to_q(input: Union[str, int]) -> Optional[ElementModQ]:
    """
    Given a Python integer, returns an ElementModQ.
    Returns `None` if the number is out of the allowed
    [0,Q) range.
    """
    i = int(input)
    if 0 <= i < Q:
        return ElementModQ(mpz(i))
    return None


def int_to_q_unchecked(i: Union[str, int]) -> ElementModQ:
    """
    Given a Python integer, returns an ElementModQ. Allows
    for the input to be out-of-bounds, and thus creating an invalid
    element (i.e., outside of [0,Q)). Useful for tests of it
    you're absolutely, positively, certain the input is in-bounds.
    """

    m = mpz(int(i))
    return ElementModQ(m)


def int_to_p(input: Union[str, int]) -> Optional[ElementModP]:
    """
    Given a Python integer, returns an ElementModP.
    Returns `None` if the number is out of the allowed
    [0,P) range.
    """
    i = int(input)
    if 0 <= i < P:
        return ElementModP(mpz(i))
    return None


def int_to_p_unchecked(i: Union[str, int]) -> ElementModP:
    """
    Given a Python integer, returns an ElementModP. Allows
    for the input to be out-of-bounds, and thus creating an invalid
    element (i.e., outside of [0,P)). Useful for tests or if
    you're absolutely, positively, certain the input is in-bounds.
    """
    m = mpz(int(i))
    return ElementModP(m)


def q_to_bytes(e: ElementModQ) -> bytes:
    """
    Returns a byte sequence from the element.
    """
    return to_binary(e.elem)


def bytes_to_q(b: bytes) -> ElementModQ:
    """
    Returns an element from a byte sequence.
    """
    return ElementModQ(mpz(from_binary(b)))


def add_q(*elems: ElementModQorInt) -> ElementModQ:
    """
    Adds together one or more elements in Q, returns the sum mod Q.
    """
    t = mpz(0)
    for e in elems:
        if isinstance(e, int):
            e = int_to_q_unchecked(e)
        t = (t + e.elem) % Q

    return ElementModQ(t)


def a_minus_b_q(a: ElementModQorInt, b: ElementModQorInt) -> ElementModQ:
    """
    Computes (a-b) mod q.
    """
    if isinstance(a, int):
        a = int_to_q_unchecked(a)
    if isinstance(b, int):
        b = int_to_q_unchecked(b)

    return ElementModQ((a.elem - b.elem) % Q)


def div_p(a: ElementModPOrQorInt, b: ElementModPOrQorInt) -> ElementModP:
    """
    Computes a/b mod p
    """
    if isinstance(a, int):
        a = int_to_p_unchecked(a)
    if isinstance(b, int):
        b = int_to_p_unchecked(b)

    inverse = invert(b.elem, mpz(P))
    return mult_p(a, int_to_p_unchecked(inverse))


def div_q(a: ElementModPOrQorInt, b: ElementModPOrQorInt) -> ElementModQ:
    """
    Computes a/b mod q
    """
    if isinstance(a, int):
        a = int_to_p_unchecked(a)
    if isinstance(b, int):
        b = int_to_p_unchecked(b)

    inverse = invert(b.elem, mpz(Q))
    return mult_q(a, int_to_q_unchecked(inverse))


def negate_q(a: ElementModQorInt) -> ElementModQ:
    """
    Computes (Q - a) mod q.
    """
    if isinstance(a, int):
        a = int_to_q_unchecked(a)
    return ElementModQ(Q - a.elem)


def a_plus_bc_q(
    a: ElementModQorInt, b: ElementModQorInt, c: ElementModQorInt
) -> ElementModQ:
    """
    Computes (a + b * c) mod q.
    """
    if isinstance(a, int):
        a = int_to_q_unchecked(a)
    if isinstance(b, int):
        b = int_to_q_unchecked(b)
    if isinstance(c, int):
        c = int_to_q_unchecked(c)

    return ElementModQ((a.elem + b.elem * c.elem) % Q)


def mult_inv_p(e: ElementModPOrQorInt) -> ElementModP:
    """
    Computes the multiplicative inverse mod p.

    :param e:  An element in [1, P).
    """
    if isinstance(e, int):
        e = int_to_p_unchecked(e)

    assert e.elem != 0, "No multiplicative inverse for zero"
    return ElementModP(powmod(e.elem, -1, P))


def pow_p(b: ElementModPOrQorInt, e: ElementModPOrQorInt) -> ElementModP:
    """
    Computes b^e mod p.

    :param b: An element in [0,P).
    :param e: An element in [0,P).
    """

    if isinstance(b, int):
        b = int_to_p_unchecked(b)
    if isinstance(e, int):
        e = int_to_p_unchecked(e)

    return ElementModP(powmod(b.elem, e.elem, P))


def pow_q(b: ElementModQorInt, e: ElementModQorInt) -> ElementModQ:
    """
    Computes b^e mod q.

    :param b: An element in [0,Q).
    :param e: An element in [0,Q).
    """
    if isinstance(b, int):
        b = int_to_q_unchecked(b)

    if isinstance(e, int):
        e = int_to_q_unchecked(e)

    return ElementModQ(powmod(b.elem, e.elem, Q))


def mult_p(*elems: ElementModPOrQorInt) -> ElementModP:
    """
    Computes the product, mod p, of all elements.

    :param elems: Zero or more elements in [0,P).
    """
    product = mpz(1)
    for x in elems:
        if isinstance(x, int):
            x = int_to_p_unchecked(x)
        product = (product * x.elem) % P
    return ElementModP(product)


def mult_q(*elems: ElementModPOrQorInt) -> ElementModQ:
    """
    Computes the product, mod q, of all elements.

    :param elems: Zero or more elements in [0,Q).
    """
    product = mpz(1)
    for x in elems:
        if isinstance(x, int):
            x = int_to_p_unchecked(x)
        product = (product * x.elem) % Q
    return ElementModQ(product)


def g_pow_p(e: ElementModPOrQ) -> ElementModP:
    """
    Computes g^e mod p.

    :param e: An element in [0,P).
    """
    return pow_p(ElementModP(mpz(G)), e)


def rand_q() -> ElementModQ:
    """
    Generate random number between 0 and Q

    :return: Random value between 0 and Q
    """
    return int_to_q_unchecked(randbelow(Q))


def rand_range_q(start: ElementModQorInt) -> ElementModQ:
    """
    Generate random number between start and Q

    :param start: Starting value of range
    :return: Random value between start and Q
    """
    if isinstance(start, ElementModQ):
        start = start.to_int()

    random = 0
    while random < start:
        random = randbelow(Q)
    return int_to_q_unchecked(random)


def eq_elems(a: ElementModPOrQ, b: ElementModPOrQ) -> bool:
    """
    Returns whether the two elements hold the same value.
    """
    return a.elem == b.elem



from typing import Dict, Iterable, List, NamedTuple, Optional
from .key_ceremony import (
    AuxiliaryPublicKey,
    CeremonyDetails,
    ElectionJointKey,
    ElectionPartialKeyBackup,
    ElectionPartialKeyChallenge,
    ElectionPartialKeyVerification,
    ElectionPublicKey,
    PublicKeySet,
    combine_election_public_keys,
    verify_election_partial_key_challenge,
)
from .types import GUARDIAN_ID, MEDIATOR_ID


class GuardianPair(NamedTuple):
    """Pair of guardians involved in sharing"""

    owner_id: GUARDIAN_ID
    designated_id: GUARDIAN_ID


class BackupVerificationState(NamedTuple):
    """The state of the verifications of all guardian election partial key backups"""

    all_sent: bool = False
    all_verified: bool = False
    failed_verifications: List[GuardianPair] = []


class KeyCeremonyMediator:
    """
    KeyCeremonyMediator for assisting communication between guardians
    """

    id: MEDIATOR_ID
    ceremony_details: CeremonyDetails

    # From Guardians
    # Round 1
    _auxiliary_public_keys: Dict[GUARDIAN_ID, AuxiliaryPublicKey]
    _election_public_keys: Dict[GUARDIAN_ID, ElectionPublicKey]

    # Round 2
    _election_partial_key_backups: Dict[GuardianPair, ElectionPartialKeyBackup]

    # Round 3
    _election_partial_key_verifications: Dict[
        GuardianPair, ElectionPartialKeyVerification
    ]

    def __init__(self, id: MEDIATOR_ID, ceremony_details: CeremonyDetails):
        self.id = id
        self.ceremony_details = ceremony_details
        self._auxiliary_public_keys: Dict[GUARDIAN_ID, AuxiliaryPublicKey] = {}
        self._election_public_keys: Dict[GUARDIAN_ID, ElectionPublicKey] = {}
        self._election_partial_key_backups: Dict[
            GuardianPair, ElectionPartialKeyBackup
        ] = {}
        self._election_partial_key_verifications: Dict[
            GuardianPair, ElectionPartialKeyVerification
        ] = {}
        self._election_partial_key_challenges: Dict[
            GuardianPair, ElectionPartialKeyChallenge
        ] = {}

    # ROUND 1: Announce guardians with public keys
    def announce(self, public_key_set: PublicKeySet) -> None:
        """
        Announce the guardian as present and participating the Key Ceremony
        :param public_key_set: Both the guardians public keys
        """
        self._receive_election_public_key(public_key_set.election)
        self._receive_auxiliary_public_key(public_key_set.auxiliary)

    def all_guardians_announced(self) -> bool:
        """
        Check the annoucement of all the guardians expected
        :return: True if all guardians in attendance are announced
        """
        return (
            self._all_auxiliary_public_keys_available()
            and self._all_election_public_keys_available()
        )

    def share_announced(
        self, requesting_guardian_id: Optional[GUARDIAN_ID] = None
    ) -> Optional[List[PublicKeySet]]:
        """
        When all guardians have announced, share their public keys indicating their announcement
        """
        if not self.all_guardians_announced():
            return None

        guardian_keys: List[PublicKeySet] = []
        for guardian_id in self._get_announced_guardians():
            if guardian_id is not requesting_guardian_id:
                guardian_keys.append(
                    PublicKeySet(
                        self._election_public_keys[guardian_id],
                        self._auxiliary_public_keys[guardian_id],
                    )
                )
        return guardian_keys

    # ROUND 2: Share Election Partial Key Backups for compensating
    def receive_backups(self, backups: List[ElectionPartialKeyBackup]) -> None:
        """
        Receive all the election partial key backups generated by a guardian
        """
        if not self.all_guardians_announced():
            return
        for backup in backups:
            self._receive_election_partial_key_backup(backup)

    def all_backups_available(self) -> bool:
        """
        Check the availability of all the guardians backups
        :return: True if all guardians have sent backups
        """
        return (
            self.all_guardians_announced()
            and self._all_election_partial_key_backups_available()
        )

    def share_backups(
        self, requesting_guardian_id: Optional[GUARDIAN_ID] = None
    ) -> Optional[List[ElectionPartialKeyBackup]]:
        """
        Share all backups designated for a specific guardian
        """
        if not self.all_guardians_announced() or not self.all_backups_available:
            return None
        if not requesting_guardian_id:
            return list(self._election_partial_key_backups.values())
        return self._share_election_partial_key_backups_to_guardian(
            requesting_guardian_id
        )

    # ROUND 3: Share verifications of backups
    def receive_backup_verifications(
        self, verifications: List[ElectionPartialKeyVerification]
    ) -> None:
        """
        Receive all the election partial key verifications performed by a guardian
        """
        if not self.all_backups_available():
            return
        for verification in verifications:
            self._receive_election_partial_key_verification(verification)

    def get_verification_state(self) -> BackupVerificationState:
        if (
            not self.all_backups_available()
            or not self._all_election_partial_key_verifications_received()
        ):
            return BackupVerificationState()
        return self._check_verification_of_election_partial_key_backups()

    def all_backups_verified(self) -> bool:
        return self.get_verification_state().all_verified

    # ROUND 4 (Optional): If a verification fails, guardian must issue challenge
    def verify_challenge(
        self, challenge: ElectionPartialKeyChallenge
    ) -> ElectionPartialKeyVerification:
        """
        Mediator receives challenge and will act to mediate and verify
        """
        verification = verify_election_partial_key_challenge(self.id, challenge)
        if verification.verified:
            self._receive_election_partial_key_verification(verification)
        return verification

    # FINAL: Publish joint public election key
    def publish_joint_key(self) -> Optional[ElectionJointKey]:
        """
        Publish joint election key from the public keys of all guardians
        :return: Joint key for election
        """
        if not self.all_backups_verified():
            return None

        return combine_election_public_keys(list(self._election_public_keys.values()))

    def reset(self, ceremony_details: CeremonyDetails) -> None:
        """
        Reset mediator to initial state
        :param ceremony_details: Ceremony details of election
        """
        self.ceremony_details = ceremony_details
        self._auxiliary_public_keys = {}
        self._election_public_keys = {}
        self._election_partial_key_backups = {}
        self._election_partial_key_challenges = {}
        self._election_partial_key_verifications = {}

    # Auxiliary Public Keys
    def _receive_auxiliary_public_key(self, public_key: AuxiliaryPublicKey) -> None:
        """
        Receive auxiliary public key from guardian
        :param public_key: Auxiliary public key
        """
        self._auxiliary_public_keys[public_key.owner_id] = public_key

    def _all_auxiliary_public_keys_available(self) -> bool:
        """
        True if all auxiliary public key for all guardians available
        :return: All auxiliary public backups for all guardians available
        """
        return (
            len(self._auxiliary_public_keys)
            == self.ceremony_details.number_of_guardians
        )

    # Election Public Keys
    def _receive_election_public_key(self, public_key: ElectionPublicKey) -> None:
        """
        Receive election public key from guardian
        :param public_key: election public key
        """
        self._election_public_keys[public_key.owner_id] = public_key

    def _all_election_public_keys_available(self) -> bool:
        """
        True if all election public keys for all guardians available
        :return: All election public keys for all guardians available
        """
        return (
            len(self._election_public_keys) == self.ceremony_details.number_of_guardians
        )

    def _get_announced_guardians(self) -> Iterable[GUARDIAN_ID]:
        return self._election_public_keys.keys()

    # Election Partial Key Backups
    def _receive_election_partial_key_backup(
        self, backup: ElectionPartialKeyBackup
    ) -> None:
        """
        Receive election partial key backup from guardian
        :param backup: Election partial key backup
        :return: boolean indicating success or failure
        """
        if backup.owner_id == backup.designated_id:
            return
        self._election_partial_key_backups[
            GuardianPair(backup.owner_id, backup.designated_id)
        ] = backup

    def _all_election_partial_key_backups_available(self) -> bool:
        """
        True if all election partial key backups for all guardians available
        :return: All election partial key backups for all guardians available
        """
        required_backups_per_guardian = self.ceremony_details.number_of_guardians - 1
        return (
            len(self._election_partial_key_backups)
            == required_backups_per_guardian * self.ceremony_details.number_of_guardians
        )

    def _share_election_partial_key_backups_to_guardian(
        self, guardian_id: GUARDIAN_ID
    ) -> List[ElectionPartialKeyBackup]:
        """
        Share all election partial key backups for designated guardian
        :param guardian_id: Recipients guardian id
        :return: List of guardians designated backups
        """
        backups: List[ElectionPartialKeyBackup] = []
        for current_guardian_id in self._get_announced_guardians():
            if guardian_id != current_guardian_id:
                backup = self._election_partial_key_backups[
                    GuardianPair(current_guardian_id, guardian_id)
                ]
                if backup is not None:
                    backups.append(backup)
        return backups

    # Partial Key Verifications
    def _receive_election_partial_key_verification(
        self, verification: ElectionPartialKeyVerification
    ) -> None:
        """
        Receive election partial key verification from guardian
        :param verification: Election partial key verification
        """
        if verification.owner_id == verification.designated_id:
            return
        self._election_partial_key_verifications[
            GuardianPair(verification.owner_id, verification.designated_id)
        ] = verification

    def _all_election_partial_key_verifications_received(self) -> bool:
        """
        True if all election partial key verifications recieved
        :return: All election partial key verifications received
        """
        required_verifications_per_guardian = (
            self.ceremony_details.number_of_guardians - 1
        )
        return (
            len(self._election_partial_key_verifications)
            == required_verifications_per_guardian
            * self.ceremony_details.number_of_guardians
        )

    def _check_verification_of_election_partial_key_backups(
        self,
    ) -> BackupVerificationState:
        """
        True if all election partial key backups verified
        :return: All election partial key backups verified
        """
        if not self._all_election_partial_key_verifications_received():
            return BackupVerificationState()
        failed_verifications: List[GuardianPair] = []
        for verification in self._election_partial_key_verifications.values():
            if not verification.verified:
                failed_verifications.append(
                    GuardianPair(verification.owner_id, verification.designated_id)
                )

        return BackupVerificationState(
            True, len(failed_verifications) == 0, failed_verifications
        )



from typing import Dict, Optional, Tuple

from .ballot import SubmittedBallot, CiphertextContest, CiphertextSelection
from .decryption_share import (
    CiphertextDecryptionSelection,
    DecryptionShare,
    get_shares_for_selection,
)
from .dlog import discrete_log
from .group import ElementModP, ElementModQ, mult_p, div_p
from .tally import (
    CiphertextTally,
    PlaintextTally,
    PlaintextTallyContest,
    PlaintextTallySelection,
)
from .logs import log_warning
from .types import CONTEST_ID, GUARDIAN_ID, SELECTION_ID

AVAILABLE_GUARDIAN_ID = GUARDIAN_ID
MISSING_GUARDIAN_ID = GUARDIAN_ID

GUARDIAN_PUBLIC_KEY = ElementModP

ELECTION_PUBLIC_KEY = ElementModP

# The methods in this file can be used to decrypt values if private keys or nonces are not known
# and the key ceremony is used to share secrets among a quorum of guardians


def decrypt_tally(
    tally: CiphertextTally,
    shares: Dict[GUARDIAN_ID, DecryptionShare],
    crypto_extended_base_hash: ElementModQ,
) -> Optional[PlaintextTally]:
    """
    Try to decrypt the tally and the spoiled ballots using the provided decryption shares.

    :param tally: The CiphertextTally to decrypt
    :param shares: The guardian Decryption Shares for all guardians
    :param context: the Ciphertextelectioncontext
    :return: A PlaintextTally or None if there is an error
    """
    contests: Dict[CONTEST_ID, PlaintextTallyContest] = {}

    for contest in tally.contests.values():
        plaintext_contest = decrypt_contest_with_decryption_shares(
            CiphertextContest(
                contest.object_id,
                contest.description_hash,
                list(contest.selections.values()),
            ),
            shares,
            crypto_extended_base_hash,
        )
        if not plaintext_contest:
            log_warning(f"contest: {contest.object_id} failed to decrypt with shares")
            return None
        contests[contest.object_id] = plaintext_contest

    return PlaintextTally(tally.object_id, contests)


def decrypt_ballot(
    ballot: SubmittedBallot,
    shares: Dict[AVAILABLE_GUARDIAN_ID, DecryptionShare],
    crypto_extended_base_hash: ElementModQ,
) -> Optional[PlaintextTally]:
    """
    Try to decrypt a single ballot using the provided decryption shares.

    :param ballot: The SubmittedBallot to decrypt
    :param shares: The guardian Decryption Shares for all guardians
    :param crypto_extended_base_hash: The extended base hash
    :return: A PlaintextTally or None if there is an error
    """
    contests: Dict[CONTEST_ID, PlaintextTallyContest] = {}

    for contest in ballot.contests:
        plaintext_contest = decrypt_contest_with_decryption_shares(
            CiphertextContest(
                contest.object_id,
                contest.description_hash,
                contest.ballot_selections,
            ),
            shares,
            crypto_extended_base_hash,
        )
        if not plaintext_contest:
            log_warning(f"contest: {contest.object_id} failed to decrypt with shares")
            return None
        contests[contest.object_id] = plaintext_contest

    return PlaintextTally(ballot.object_id, contests)


def decrypt_contest_with_decryption_shares(
    contest: CiphertextContest,
    shares: Dict[GUARDIAN_ID, DecryptionShare],
    crypto_extended_base_hash: ElementModQ,
) -> Optional[PlaintextTallyContest]:
    """
    Decrypt the specified contest within the context of the specified Decryption Shares.

    :param contest: the contest to decrypt
    :param shares: a collection of `DecryptionShare` used to decrypt
    :param crypto_extended_base_hash: the extended base hash code (𝑄') for the election
    :return: a collection of `PlaintextTallyContest` or `None` if there is an error
    """
    plaintext_selections: Dict[SELECTION_ID, PlaintextTallySelection] = {}

    for selection in contest.selections:
        tally_shares = get_shares_for_selection(selection.object_id, shares)
        plaintext_selection = decrypt_selection_with_decryption_shares(
            selection, tally_shares, crypto_extended_base_hash
        )
        if plaintext_selection is None:
            log_warning(
                (
                    f"could not decrypt contest {contest.object_id} "
                    f"with selection {selection.object_id}"
                )
            )
            return None
        plaintext_selections[plaintext_selection.object_id] = plaintext_selection

    return PlaintextTallyContest(contest.object_id, plaintext_selections)


def decrypt_selection_with_decryption_shares(
    selection: CiphertextSelection,
    shares: Dict[
        GUARDIAN_ID, Tuple[ELECTION_PUBLIC_KEY, CiphertextDecryptionSelection]
    ],
    crypto_extended_base_hash: ElementModQ,
    suppress_validity_check: bool = False,
) -> Optional[PlaintextTallySelection]:
    """
    Decrypt the specified `CiphertextSelection` with the collection of `ElementModP` decryption shares.
    Each share is expected to be passed with the corresponding public key so that the encryption can be validated

    :param selection: a `CiphertextSelection`
    :param shares: the collection of shares to decrypt the selection
    :param crypto_extended_base_hash: the extended base hash code (𝑄') for the election
    :param suppress_validity_check: do not validate the encryption prior to decrypting (useful for tests)
    :return: a `PlaintextTallySelection` or `None` if there is an error
    """
    if not suppress_validity_check:
        # Verify that all of the shares are computed correctly
        for share in shares.values():
            public_key, decryption = share
            # verify we have a proof or recovered parts
            if not decryption.is_valid(
                selection.ciphertext, public_key, crypto_extended_base_hash
            ):
                log_warning(
                    f"share: {decryption.object_id} has invalid proof or recovered parts"
                )
                return None

    # accumulate all of the shares calculated for the selection
    all_shares_product_M = mult_p(
        *[decryption.share for (_, decryption) in shares.values()]
    )

    # Calculate 𝑀=𝐵⁄(∏𝑀𝑖) mod 𝑝.
    decrypted_value = div_p(selection.ciphertext.data, all_shares_product_M)
    d_log = discrete_log(decrypted_value)
    return PlaintextTallySelection(
        selection.object_id,
        d_log,
        decrypted_value,
        selection.ciphertext,
        [share for (guardian_id, (public_key, share)) in shares.items()],
    )



from typing import List, NamedTuple

from .elgamal import ElGamalKeyPair
from .group import (
    add_q,
    ElementModP,
    ElementModQ,
    g_pow_p,
    int_to_p_unchecked,
    div_q,
    mult_p,
    mult_q,
    ONE_MOD_P,
    pow_p,
    pow_q,
    rand_q,
    ZERO_MOD_Q,
    Q,
)
from .schnorr import make_schnorr_proof, SchnorrProof

SECRET_COEFFICIENT = ElementModQ  # Secret coefficient of election polynomial
PUBLIC_COMMITMENT = ElementModP  # Public commitment of election polynomial

# TODO:ISSUE #84: do not use lists here
class ElectionPolynomial(NamedTuple):
    """
    A polynomial defined by coefficients

    The 0-index coefficient is used for a secret key which can
    be discovered by a quorum of n guardians corresponding to n coefficients.
    """

    coefficients: List[SECRET_COEFFICIENT]
    """The secret coefficients `a_ij` """

    coefficient_commitments: List[PUBLIC_COMMITMENT]
    """The public keys `K_ij`generated from secret coefficients"""

    coefficient_proofs: List[SchnorrProof]
    """A proof of posession of the private key for each secret coefficient"""


def generate_polynomial(
    number_of_coefficients: int, nonce: ElementModQ = None
) -> ElectionPolynomial:
    """
    Generates a polynomial for sharing election keys

    :param number_of_coefficients: Number of coefficients of polynomial
    :param nonce: an optional nonce parameter that may be provided (useful for testing)
    :return: Polynomial used to share election keys
    """
    coefficients: List[SECRET_COEFFICIENT] = []
    commitments: List[PUBLIC_COMMITMENT] = []
    proofs: List[SchnorrProof] = []

    for i in range(number_of_coefficients):
        # Note: the nonce value is not safe.  it is designed for testing only.
        # this method should be called without the nonce in production.
        coefficient = add_q(nonce, i) if nonce is not None else rand_q()
        commitment = g_pow_p(coefficient)
        proof = make_schnorr_proof(
            ElGamalKeyPair(coefficient, commitment), rand_q()
        )  # TODO Alternate schnoor proof method that doesn't need KeyPair

        coefficients.append(coefficient)
        commitments.append(commitment)
        proofs.append(proof)
    return ElectionPolynomial(coefficients, commitments, proofs)


def compute_polynomial_coordinate(
    exponent_modifier: int, polynomial: ElectionPolynomial
) -> ElementModQ:
    """
    Computes a single coordinate value of the election polynomial used for sharing

    :param exponent_modifier: Unique modifier (usually sequence order) for exponent
    :param polynomial: Election polynomial
    :return: Polynomial used to share election keys
    """

    _check_exponent_modifier(exponent_modifier)

    computed_value = ZERO_MOD_Q
    for (i, coefficient) in enumerate(polynomial.coefficients):
        exponent = pow_q(exponent_modifier, i)
        factor = mult_q(coefficient, exponent)
        computed_value = add_q(computed_value, factor)
    return computed_value


# pylint: disable=unnecessary-comprehension
def compute_lagrange_coefficient(coordinate: int, *degrees: int) -> ElementModQ:
    """
    Compute the lagrange coefficient for a specific coordinate against N degrees.
    :param coordinate: the coordinate to plot, uisually a Guardian's Sequence Order
    :param degrees: the degrees across which to plot, usually the collection of
                    available Guardians' Sequence Orders
    """

    numerator = mult_q(*[degree for degree in degrees])
    denominator = mult_q(*[(degree - coordinate) for degree in degrees])
    result = div_q((numerator), (denominator))
    return result


def verify_polynomial_coordinate(
    coordinate: ElementModQ,
    exponent_modifier: int,
    commitments: List[PUBLIC_COMMITMENT],
) -> bool:
    """
    Verify a polynomial coordinate value is in fact on the polynomial's curve

    :param coordinate: Value to be checked
    :param exponent_modifier: Unique modifier (usually sequence order) for exponent
    :param commitments: Public commitments for coefficients of polynomial
    :return: True if verified on polynomial
    """

    _check_exponent_modifier(exponent_modifier)

    commitment_output = ONE_MOD_P
    for (i, commitment) in enumerate(commitments):
        exponent = pow_p(int_to_p_unchecked(exponent_modifier), int_to_p_unchecked(i))
        factor = pow_p(commitment, exponent)
        commitment_output = mult_p(commitment_output, factor)

    value_output = g_pow_p(coordinate)
    return value_output == commitment_output


def _check_exponent_modifier(exponent_modifier: int) -> None:
    """
    Ensure the exponent modifier (typically sequence order)
    is between 0 and Q. If not, identify modifier is out of the range
    """
    assert 0 < exponent_modifier < Q, "exponent_modifier is out of range"



# pylint: disable=consider-using-with
from __future__ import annotations
import traceback
from typing import Any, Callable, Iterable, List, TypeVar
from contextlib import AbstractContextManager
from multiprocessing import Pool as ProcessPool
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing.pool import Pool
from psutil import cpu_count

from .logs import log_warning
from .singleton import Singleton

_T = TypeVar("_T")


class Scheduler(Singleton, AbstractContextManager):
    """
    Worker that wraps Multprocessing and allows
    for shared context or spawning processes.
    Implemented as a singleton to guarantee there is only one set
    of tread and process pools in use throughout the library.
    Also implements the [Context Manager Protocol](https://docs.python.org/3.8/library/stdtypes.html#typecontextmanager)
    """

    __process_pool: Pool
    __thread_pool: Pool

    def __init__(self) -> None:
        super().__init__()
        self._open()

    def __enter__(self) -> Scheduler:
        self._open()
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        self.close()

    def _open(self) -> None:
        """Open pools"""
        max_processes = cpu_count(logical=False)
        # Reserve one CPU for I/O bound tasks
        if max_processes > 2:
            max_processes = max_processes - 1
        self.__process_pool = ProcessPool(max_processes)
        self.__thread_pool = ThreadPool(max_processes)

    def close(self) -> None:
        """Close pools"""
        self.__process_pool.close()
        self.__thread_pool.close()

    @staticmethod
    def cpu_count() -> int:
        """Get CPU count"""
        return int(cpu_count(logical=False))

    def schedule(
        self,
        task: Callable,
        arguments: Iterable[Iterable[Any]],
        with_shared_resources: bool = False,
    ) -> List[_T]:
        """
        Schedule tasks with list of arguments
        :param task: the callable task to execute
        :param arguments: the list of lists passed to the task using starmap
        :param with_shared_resources: flag to use threads instead of processes
            allowing resources to be shared.  note
            when using the threadpool, execution is bound
            by default to the [global interpreter lock]
            (https://docs.python.org/3.8/glossary.html#term-global-interpreter-lock)
        """
        if with_shared_resources:
            return self.safe_starmap(self.__thread_pool, task, arguments)
        return self.safe_starmap(self.__process_pool, task, arguments)

    @staticmethod
    def safe_starmap(
        pool: Pool, task: Callable, arguments: Iterable[Iterable[Any]]
    ) -> List[_T]:
        """Safe wrapper around starmap to ensure pool is open"""
        try:
            return pool.starmap(task, arguments)
        except ValueError as e:
            log_warning(
                f"safe_starmap({task}, {arguments}) exception ValueError({str(e)})"
            )
            return []
        except Exception:  # pylint: disable=broad-except
            log_warning(
                f"safe_starmap({task}, {arguments}) failed with \n {traceback.format_exc()}"
            )
            return []

    @staticmethod
    def safe_map(pool: Pool, task: Callable, arguments: Iterable[Any]) -> List[_T]:
        """Safe wrapper around starmap to ensure pool is open"""
        try:
            return pool.map(task, arguments)
        except ValueError as e:
            log_warning(f"safe_map({task}, {arguments}) exception ValueError({str(e)})")
            return []
        except Exception:  # pylint: disable=broad-except
            log_warning(
                f"safe_starmap({task}, {arguments}) failed with \n {traceback.format_exc()}"
            )
            return []



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Tuple

from electionguard.key_ceremony import ELECTION_JOINT_PUBLIC_KEY

from .election import CiphertextElectionContext, make_ciphertext_election_context
from .group import ElementModQ
from .manifest import Manifest, InternalManifest
from .utils import get_optional


@dataclass
class ElectionBuilder:
    """
    `ElectionBuilder` is a stateful builder object that constructs `CiphertextElectionContext` objects
    following the initialization process that ElectionGuard Expects.
    SUGGESTION Perhaps move to test package.
    """

    number_of_guardians: int
    """
    The number of guardians necessary to generate the public key
    """
    quorum: int
    """
    The quorum of guardians necessary to decrypt an election.  Must be less than `number_of_guardians`
    """

    manifest: Manifest

    internal_manifest: InternalManifest = field(init=False)

    election_joint_public_key: Optional[ELECTION_JOINT_PUBLIC_KEY] = field(default=None)

    commitment_hash: Optional[ElementModQ] = field(default=None)

    def __post_init__(self) -> None:
        self.internal_manifest = InternalManifest(self.manifest)

    def set_public_key(
        self, election_joint_public_key: ELECTION_JOINT_PUBLIC_KEY
    ) -> ElectionBuilder:
        """
        Set election public key
        :param election_joint_public_key: elgamal public key for election
        :return: election builder
        """
        self.election_joint_public_key = election_joint_public_key
        return self

    def set_commitment_hash(self, commitment_hash: ElementModQ) -> ElectionBuilder:
        """
        Set commitment hash
        :param commitment_hash: hash of the commitments guardians make to each other
        :return: election builder
        """
        self.commitment_hash = commitment_hash
        return self

    def build(
        self,
    ) -> Optional[Tuple[InternalManifest, CiphertextElectionContext]]:
        """
        Build election
        :return: election manifest and context or none
        """
        if not self.manifest.is_valid():
            return None

        if self.election_joint_public_key is None:
            return None

        return (
            self.internal_manifest,
            make_ciphertext_election_context(
                self.number_of_guardians,
                self.quorum,
                get_optional(self.election_joint_public_key),
                get_optional(self.commitment_hash),
                self.manifest.crypto_hash(),
            ),
        )



from typing import Dict, List, Optional

from .ballot import SubmittedBallot
from .decryption import (
    compute_lagrange_coefficients_for_guardians,
    reconstruct_decryption_share,
    reconstruct_decryption_share_for_ballot,
)
from .decryption_share import DecryptionShare, CompensatedDecryptionShare
from .decrypt_with_shares import decrypt_ballot, decrypt_tally
from .election import CiphertextElectionContext
from .key_ceremony import ElectionPublicKey
from .key_ceremony_mediator import GuardianPair
from .logs import log_info, log_warning
from .tally import (
    CiphertextTally,
    PlaintextTally,
)
from .types import BALLOT_ID, GUARDIAN_ID, MEDIATOR_ID


class DecryptionMediator:
    """
    The Decryption Mediator composes partial decryptions from each Guardian
    to form a decrypted representation of an election tally.
    """

    # pylint: disable=too-many-instance-attributes
    id: MEDIATOR_ID
    _context: CiphertextElectionContext

    # Guardians
    _available_guardians: Dict[GUARDIAN_ID, ElectionPublicKey]
    _missing_guardians: Dict[GUARDIAN_ID, ElectionPublicKey]

    # Decryption Shares
    _tally_shares: Dict[GUARDIAN_ID, DecryptionShare]
    _ballot_shares: Dict[BALLOT_ID, Dict[GUARDIAN_ID, DecryptionShare]]

    # Compensated Shares
    _compensated_tally_shares: Dict[GuardianPair, CompensatedDecryptionShare]
    _compensated_ballot_shares: Dict[
        BALLOT_ID, Dict[GuardianPair, CompensatedDecryptionShare]
    ]

    def __init__(self, id: MEDIATOR_ID, context: CiphertextElectionContext):
        """Initialize the decryption mediator."""
        self.id = id
        self._context = context

        self._available_guardians = {}
        self._missing_guardians = {}

        self._tally_shares = {}
        self._ballot_shares = {}

        self._compensated_tally_shares = {}
        self._compensated_ballot_shares = {}

    def announce(
        self,
        guardian_key: ElectionPublicKey,
        tally_share: DecryptionShare,
        ballot_shares: Dict[BALLOT_ID, Optional[DecryptionShare]] = None,
    ) -> None:
        """
        Announce that a Guardian is present and participating in the decryption.

        A guardian announces by presenting their id and their shares of the decryption

        :param guardian_key: The election public key of the guardian who will participate in the decryption.
        :param tally_share: Guardian's decryption share of the tally
        :param ballot_shares: Guardian's decryption shares of the ballots
        """
        guardian_id = guardian_key.owner_id

        # Only allow a guardian to announce once
        if guardian_id in self._available_guardians:
            log_info(f"guardian {guardian_id} already announced")
            return

        self._save_tally_share(guardian_id, tally_share)

        if ballot_shares is not None:
            self._save_ballot_shares(guardian_id, ballot_shares)

        self._mark_available(guardian_key)

    def announce_missing(self, missing_guardian_key: ElectionPublicKey) -> None:
        """
        Announce that a Guardian is missing and not participating in the decryption.

        :param missing_guardian_key: The election public key of the missing guardian
        """
        missing_guardian_id = missing_guardian_key.owner_id

        # If guardian is available, can't be marked missing
        if missing_guardian_id in self._available_guardians:
            log_info(f"guardian {missing_guardian_id} already announced")
            return

        self._mark_missing(missing_guardian_key)

    def validate_missing_guardians(
        self, guardian_keys: List[ElectionPublicKey]
    ) -> bool:
        """Check the guardian's collections of keys and ensure the public keys match for the guardians."""
        # Check this guardian's collection of public keys
        # for other guardians that have not announced
        missing_guardians: Dict[GUARDIAN_ID, ElectionPublicKey] = {
            guardian_key.owner_id: guardian_key
            for guardian_key in guardian_keys
            if guardian_key.owner_id not in self._available_guardians
        }

        # Check that the public keys match for any missing guardians already reported
        # note this check naively assumes that the first guardian to annouce is telling the truth
        # but for this implementation it is simply a sanity check on the input data.
        # a consuming application should implement better validation of the guardian state
        # before announcing a guardian is available for decryption.
        for guardian_id, public_key in missing_guardians.items():
            if guardian_id in self._missing_guardians.keys():
                if self._missing_guardians[guardian_id] != public_key:
                    log_warning(
                        (
                            f"announce guardian: {guardian_id} "
                            f"expected public key mismatch for missing {guardian_id}"
                        )
                    )
                    return False
            else:
                self._missing_guardians[guardian_id] = missing_guardians[guardian_id]
        return True

    def announcement_complete(self) -> bool:
        """
        Determine if the announcement phase is complete
        :return: True if announcement complete
        """
        # If a quorum not announced, not ready
        if len(self._available_guardians) < self._context.quorum:
            log_warning("cannot decrypt with less than quorum available guardians")
            return False

        # If guardians missing or available not accounted for, not ready
        if (
            len(self._available_guardians) + len(self._missing_guardians)
            != self._context.number_of_guardians
        ):
            log_warning(
                "cannot decrypt without accounting for all guardians missing or present"
            )
            return False
        return True

    def get_available_guardians(self) -> List[ElectionPublicKey]:
        """
        Get all available guardian keys
        :return: All available guardians election public keys
        """
        return list(self._available_guardians.values())

    def get_missing_guardians(self) -> List[ElectionPublicKey]:
        """
        Get all missing guardian keys
        :return: All missing guardians election public keys
        """
        return list(self._missing_guardians.values())

    def receive_tally_compensation_share(
        self, tally_compensation_share: CompensatedDecryptionShare
    ) -> None:
        self._compensated_tally_shares[
            GuardianPair(
                tally_compensation_share.guardian_id,
                tally_compensation_share.missing_guardian_id,
            )
        ] = tally_compensation_share

    def receive_ballot_compensation_shares(
        self, ballot_compensation_shares: Dict[BALLOT_ID, CompensatedDecryptionShare]
    ) -> None:
        for ballot_id, share in ballot_compensation_shares.items():
            ballot_shares = self._compensated_ballot_shares.get(ballot_id)
            if not ballot_shares:
                ballot_shares = {}
            ballot_shares[
                GuardianPair(share.guardian_id, share.missing_guardian_id)
            ] = share
            self._compensated_ballot_shares[ballot_id] = ballot_shares

    def reconstruct_shares_for_tally(self, ciphertext_tally: CiphertextTally) -> None:
        lagrange_coefficients = compute_lagrange_coefficients_for_guardians(
            list(self._available_guardians.values())
        )
        for (
            missing_guardian_id,
            missing_guardian_key,
        ) in self._missing_guardians.items():
            # Share already reconstructed
            if missing_guardian_id in self._tally_shares:
                continue

            compensated_shares = _filter_by_missing_guardian(
                missing_guardian_id, self._compensated_tally_shares
            )

            reconstructed_share = reconstruct_decryption_share(
                missing_guardian_key,
                ciphertext_tally,
                compensated_shares,
                lagrange_coefficients,
            )

            if reconstruct_decryption_share is None:
                log_warning(
                    f"failed to reconstruct tally share for missing guardian {missing_guardian_id}"
                )

            # Add reconstructed share into tally shares
            self._tally_shares[missing_guardian_id] = reconstructed_share

    def reconstruct_shares_for_ballots(
        self, ciphertext_ballots: List[SubmittedBallot]
    ) -> None:
        lagrange_coefficients = compute_lagrange_coefficients_for_guardians(
            list(self._available_guardians.values())
        )
        for ciphertext_ballot in ciphertext_ballots:
            ballot_id = ciphertext_ballot.object_id
            ballot_shares = self._ballot_shares[ballot_id]

            for (
                missing_guardian_id,
                missing_guardian_key,
            ) in self._missing_guardians.items():
                # Share already reconstructed
                if missing_guardian_id in ballot_shares:
                    continue

                compensated_shares = _filter_by_missing_guardian(
                    missing_guardian_id, self._compensated_ballot_shares[ballot_id]
                )

                reconstructed_share = reconstruct_decryption_share_for_ballot(
                    missing_guardian_key,
                    ciphertext_ballot,
                    compensated_shares,
                    lagrange_coefficients,
                )

                if reconstructed_share is None:
                    log_warning(
                        f"failed to reconstruct ballot share for {ballot_id} for missing guardian {missing_guardian_id}"
                    )

                ballot_shares[missing_guardian_id] = reconstructed_share

            # Add shares into ballot shares
            self._ballot_shares[ballot_id] = ballot_shares

    def get_plaintext_tally(
        self, ciphertext_tally: CiphertextTally
    ) -> Optional[PlaintextTally]:
        """
        Get the plaintext tally for the election by composing each Guardian's
        decrypted representation of each selection into a decrypted representation

        :return: a `PlaintextTally` or `None`
        """

        if not self.announcement_complete() or not self._ready_to_decrypt(
            self._tally_shares
        ):
            return None

        return decrypt_tally(
            ciphertext_tally,
            self._tally_shares,
            self._context.crypto_extended_base_hash,
        )

    def get_plaintext_ballots(
        self, ciphertext_ballots: List[SubmittedBallot]
    ) -> Optional[Dict[BALLOT_ID, PlaintextTally]]:
        """
        Get the plaintext ballots for the election by composing each Guardian's
        decrypted representation of each selection into a decrypted representation
        This is typically used in the spoiled ballot use case.

        :return: a Plaintext Ballots or `None`
        """

        if not self.announcement_complete():
            return None

        ballots = {}

        for ciphertext_ballot in ciphertext_ballots:
            ballot_shares = self._ballot_shares.get(ciphertext_ballot.object_id)
            if not ballot_shares or not self._ready_to_decrypt(ballot_shares):
                # Skip ballot if not ready to decrypt
                continue
            ballot = decrypt_ballot(
                ciphertext_ballot,
                ballot_shares,
                self._context.crypto_extended_base_hash,
            )

            if ballot:
                ballots[ballot.object_id] = ballot

        return ballots

    def _save_tally_share(
        self, guardian_id: GUARDIAN_ID, guardians_tally_share: DecryptionShare
    ) -> None:
        """Save a guardians tally share."""
        self._tally_shares[guardian_id] = guardians_tally_share

    def _save_ballot_shares(
        self,
        guardian_id: GUARDIAN_ID,
        guardians_ballot_shares: Dict[BALLOT_ID, Optional[DecryptionShare]],
    ) -> None:
        """Save a guardian's set of ballot shares."""
        for ballot_id, guardian_ballot_share in guardians_ballot_shares.items():
            shares = self._ballot_shares.get(ballot_id)
            if shares is None:
                shares = {}
            if guardian_ballot_share is not None:
                shares[guardian_id] = guardian_ballot_share
            self._ballot_shares[ballot_id] = shares

    def _mark_available(self, guardian_key: ElectionPublicKey) -> None:
        """
        This guardian removes itself from the missing list since it generated a valid share.
        """
        guardian_id = guardian_key.owner_id
        self._available_guardians[guardian_id] = guardian_key
        if guardian_id in self._missing_guardians:
            self._missing_guardians.pop(guardian_id)

    def _mark_missing(self, guardian_key: ElectionPublicKey) -> None:
        """"""
        self._missing_guardians[guardian_key.owner_id] = guardian_key

    def _ready_to_decrypt(self, shares: Dict[GUARDIAN_ID, DecryptionShare]) -> bool:
        """Shares are ready to decrypt."""
        # If all guardian shares are represented including if necessary
        # the missing guardians reconstructed shares, the decryption can be made
        return len(shares) == self._context.number_of_guardians


def _filter_by_missing_guardian(
    missing_guardian_id: GUARDIAN_ID,
    shares: Dict[GuardianPair, CompensatedDecryptionShare],
) -> Dict[GUARDIAN_ID, CompensatedDecryptionShare]:
    """
    Filter a guardian pair and compensated share dictionary by missing guardian.
    """
    missing_guardian_shares = {}
    for pair, share in shares.items():
        if pair.designated_id is missing_guardian_id:
            missing_guardian_shares[pair.owner_id] = share
    return missing_guardian_shares



from .ballot import CiphertextBallot, CiphertextBallotContest, CiphertextBallotSelection
from .election import CiphertextElectionContext
from .logs import log_warning
from .manifest import (
    ContestDescriptionWithPlaceholders,
    InternalManifest,
    SelectionDescription,
)


def ballot_is_valid_for_election(
    ballot: CiphertextBallot,
    internal_manifest: InternalManifest,
    context: CiphertextElectionContext,
) -> bool:
    """
    Determine if a ballot is valid for a given election
    """

    if not ballot_is_valid_for_style(ballot, internal_manifest):
        return False

    if not ballot.is_valid_encryption(
        internal_manifest.manifest_hash,
        context.elgamal_public_key,
        context.crypto_extended_base_hash,
    ):
        log_warning(
            f"ballot_is_valid_for_election: mismatching ballot encryption {ballot.object_id}"
        )
        return False

    return True


def selection_is_valid_for_style(
    selection: CiphertextBallotSelection, description: SelectionDescription
) -> bool:
    """
    Determine if selection is valid for ballot style
    :param selection: Ballot selection
    :param description: Selection description
    :return: Is valid
    """
    if selection.description_hash != description.crypto_hash():
        log_warning(
            (
                f"ballot is not valid for style: mismatched selection description hash {selection.description_hash} "
                f"for selection {description.object_id} hash {description.crypto_hash()}"
            )
        )
        return False

    return True


def contest_is_valid_for_style(
    contest: CiphertextBallotContest, description: ContestDescriptionWithPlaceholders
) -> bool:
    """
    Determine if contest is valid for ballot style
    :param contest: Contest
    :param description: Contest description
    :return: Is valid
    """
    # verify the hash matches
    if contest.description_hash != description.crypto_hash():
        log_warning(
            (
                f"ballot is not valid for style: mismatched description hash {contest.description_hash} "
                f"for contest {description.object_id} hash {description.crypto_hash()}"
            )
        )
        return False

    # verify the placeholder count
    if len(contest.ballot_selections) != len(description.ballot_selections) + len(
        description.placeholder_selections
    ):
        log_warning(
            f"ballot is not valid for style: mismatched selection count for contest {description.object_id}"
        )
        return False

    return True


def ballot_is_valid_for_style(
    ballot: CiphertextBallot, internal_manifest: InternalManifest
) -> bool:
    """
    Determine if ballot is valid for ballot style
    :param ballot: Ballot
    :param internal_manifest: Internal election description
    :return: Is valid
    """
    descriptions = internal_manifest.get_contests_for(ballot.style_id)

    for description in descriptions:
        use_contest = None
        for contest in ballot.contests:
            if description.object_id == contest.object_id:
                use_contest = contest
                break

        # verify the contest exists on the ballot
        if use_contest is None:
            log_warning(
                f"ballot is not valid for style: missing contest {description.object_id}"
            )
            return False

        if not contest_is_valid_for_style(use_contest, description):
            return False

        # verify the selection metadata
        for selection_description in description.ballot_selections:
            use_selection = None
            for selection in use_contest.ballot_selections:
                if selection_description.object_id == selection.object_id:
                    use_selection = selection
                    break

            if use_selection is None:
                log_warning(
                    f"ballot is not valid for style: missing selection {selection_description.object_id}"
                )
                return False

            if not selection_is_valid_for_style(use_selection, selection_description):
                return False

    return True



from typing import Any, Tuple
from os.path import join, dirname, realpath
from json import load
from jsonschema import validate
from jsonschema.exceptions import ValidationError

__all__ = ["get_election_description_schema", "validate_json_schema"]


def _load_schema(json_schema_file_name: str) -> Any:
    """Loads the given schema"""
    with open(join(dirname(realpath(__file__)), json_schema_file_name), "r") as file:
        schema = load(file)
    return schema


def get_election_description_schema() -> Any:
    """Get default schema for election description schema"""
    return _load_schema("election_description_schema.json")


def validate_json_schema(
    json_data: Any,
    json_schema: Any,
) -> Tuple[bool, str]:
    """Validate json schema"""
    try:
        validate(instance=json_data, schema=json_schema)
    except ValidationError as err:
        return (False, err.message)
    return (True, "Json schema validated")



import inspect
import logging
import os.path
import sys
from typing import Any, List, Tuple
from logging.handlers import RotatingFileHandler

from .singleton import Singleton

FORMAT = "[%(process)d:%(asctime)s]:%(levelname)s:%(message)s"


class ElectionGuardLog(Singleton):
    """
    A singleton log for the library
    """

    __logger: logging.Logger

    def __init__(self) -> None:
        super().__init__()

        self.__logger = logging.getLogger("electionguard")
        self.__logger.addHandler(get_stream_handler())

    @staticmethod
    def __get_call_info() -> Tuple[str, str, int]:
        stack = inspect.stack()

        # stack[0]: __get_call_info
        # stack[1]: __formatted_message
        # stack[2]: (log method, e.g. "warn")
        # stack[3]: Singleton
        # stack[4]: caller <-- we want this

        filename = stack[4][1]
        line = stack[4][2]
        funcname = stack[4][3]

        return filename, funcname, line

    def __formatted_message(self, message: str) -> str:
        filename, funcname, line = self.__get_call_info()
        message = f"{os.path.basename(filename)}.{funcname}:#L{line}: {message}"
        return message

    def add_handler(self, handler: logging.Handler) -> None:
        """
        Adds a logger handler
        """
        self.__logger.addHandler(handler)

    def remove_handler(self, handler: logging.Handler) -> None:
        """
        Removes a logger handler
        """
        self.__logger.removeHandler(handler)

    def handlers(self) -> List[logging.Handler]:
        """
        Returns all logging handlers
        """
        return self.__logger.handlers

    def debug(self, message: str, *args: Any, **kwargs: Any) -> None:
        """
        Logs a debug message
        """
        self.__logger.debug(self.__formatted_message(message), *args, **kwargs)

    def info(self, message: str, *args: Any, **kwargs: Any) -> None:
        """
        Logs a info message
        """
        self.__logger.info(self.__formatted_message(message), *args, **kwargs)

    def warn(self, message: str, *args: Any, **kwargs: Any) -> None:
        """
        Logs a warning message
        """
        self.__logger.warning(self.__formatted_message(message), *args, **kwargs)

    def error(self, message: str, *args: Any, **kwargs: Any) -> None:
        """
        Logs a error message
        """
        self.__logger.error(self.__formatted_message(message), *args, **kwargs)

    def critical(self, message: str, *args: Any, **kwargs: Any) -> None:
        """
        Logs a critical message
        """
        self.__logger.critical(self.__formatted_message(message), *args, **kwargs)


def get_stream_handler() -> logging.StreamHandler:
    """
    Get a Stream Handler, sends only warnings and errors to stdout.
    """
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(FORMAT))
    return stream_handler


def get_file_handler() -> logging.FileHandler:
    """
    Get a File System Handler, sends verbose logging to a file, `electionguard.log`.
    When that file gets too large, the logs will rotate, creating files with names
    like `electionguard.log.1`.
    """

    # TODO: add file compression, save a bunch of space.
    #   https://medium.com/@rahulraghu94/overriding-pythons-timedrotatingfilehandler-to-compress-your-log-files-iot-c766a4ace240 # pylint: disable=line-too-long
    file_handler = RotatingFileHandler(
        "electionguard.log", "a", maxBytes=10_000_000, backupCount=10
    )
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(FORMAT))
    return file_handler


LOG = ElectionGuardLog()


def log_add_handler(handler: logging.Handler) -> None:
    """
    Adds a handler to the logger
    """
    LOG.add_handler(handler)


def log_remove_handler(handler: logging.Handler) -> None:
    """
    Removes a handler from the logger
    """
    LOG.remove_handler(handler)


def log_handlers() -> List[logging.Handler]:
    """
    Returns all logger handlers
    """
    return LOG.handlers()


def log_debug(msg: str, *args: Any, **kwargs: Any) -> None:
    """
    Logs a debug message to the console and the file log.
    """
    LOG.debug(msg, *args, **kwargs)


def log_info(msg: str, *args: Any, **kwargs: Any) -> None:
    """
    Logs an information message to the console and the file log.
    """
    LOG.info(msg, *args, **kwargs)


def log_warning(msg: str, *args: Any, **kwargs: Any) -> None:
    """
    Logs a warning message to the console and the file log.
    """
    LOG.warn(msg, *args, **kwargs)


def log_error(msg: str, *args: Any, **kwargs: Any) -> None:
    """
    Logs an error message to the console and the file log.
    """
    LOG.error(msg, *args, **kwargs)


def log_critical(msg: str, *args: Any, **kwargs: Any) -> None:
    """
    Logs a critical message to the console and the file log.
    """
    LOG.critical(msg, *args, **kwargs)



from typing import Callable, Optional, NamedTuple

from .types import GUARDIAN_ID

MESSAGE = str
AUXILIARY_PUBLIC_KEY = str
AUXILIARY_SECRET_KEY = str
ENCRYPTED_MESSAGE = str


class AuxiliaryPublicKey(NamedTuple):
    """A tuple of auxiliary public key and owner information that can be shared between guardians"""

    owner_id: GUARDIAN_ID
    """
    The unique identifier of the guardian owning the key
    """

    sequence_order: int
    """
    The sequence order of the auxiliary public key (usually the guardian's sequence order)
    """

    key: AUXILIARY_PUBLIC_KEY
    """
    A string representation of the Auxiliary public key that can be shared between guardians.
    It is up to the external `AuxiliaryEncrypt` function to know how to parse this value
    """


class AuxiliaryKeyPair(NamedTuple):
    """A tuple of a secret key and public key."""

    owner_id: GUARDIAN_ID
    """
    The unique identifier of the guardian owning the key
    """

    sequence_order: int
    """
    The sequence order of the auxiliary public key (usually the guardian's sequence order)
    """

    secret_key: AUXILIARY_SECRET_KEY
    """The secret or private key"""

    public_key: AUXILIARY_PUBLIC_KEY
    """
    A string representation of the Auxiliary public key that can be shared between guardians.
    It is up to the external `AuxiliaryEncrypt` function to know how to parse this value
    """

    def share(self) -> AuxiliaryPublicKey:
        """Share the auxiliary public key and associated data"""
        return AuxiliaryPublicKey(self.owner_id, self.sequence_order, self.public_key)


AuxiliaryEncrypt = Callable[
    [MESSAGE, AUXILIARY_PUBLIC_KEY], Optional[ENCRYPTED_MESSAGE]
]
"""A callable type that represents the auxiliary encryption scheme."""

AuxiliaryDecrypt = Callable[
    [ENCRYPTED_MESSAGE, AUXILIARY_SECRET_KEY], Optional[MESSAGE]
]
"""A callable type that represents the auxiliary decryption scheme."""



from os import path
from typing import Iterable

from .ballot import PlaintextBallot, CiphertextBallot, SubmittedBallot
from .guardian import GuardianRecord, PrivateGuardianRecord
from .election import CiphertextElectionContext, ElectionConstants
from .encrypt import EncryptionDevice
from .manifest import Manifest
from .tally import PlaintextTally, PublishedCiphertextTally
from .utils import make_directory

RESULTS_DIR = "results"
MANIFEST_FILE_NAME = "manifest"
CONTEXT_FILE_NAME = "context"
CONSTANTS_FILE_NAME = "constants"
ENCRYPTED_TALLY_FILE_NAME = "encrypted_tally"
TALLY_FILE_NAME = "tally"

DEVICE_PREFIX = "device_"
COEFFICIENT_PREFIX = "coefficient_validation_set_"
BALLOT_PREFIX = "ballot_"

PLAINTEXT_BALLOT_PREFIX = "plaintext_ballot_"
GUARDIAN_PREFIX = "guardian_"

# TODO #148 Revert PlaintextTally to PublishedPlaintextTally after moving spoiled info
def publish(
    manifest: Manifest,
    context: CiphertextElectionContext,
    constants: ElectionConstants,
    devices: Iterable[EncryptionDevice],
    ciphertext_ballots: Iterable[SubmittedBallot],
    spoiled_ballots: Iterable[PlaintextTally],
    ciphertext_tally: PublishedCiphertextTally,
    plaintext_tally: PlaintextTally,
    guardian_records: Iterable[GuardianRecord],
    results_directory: str = RESULTS_DIR,
) -> None:
    """Publishes the election record as json"""

    devices_directory = path.join(results_directory, "devices")
    guardian_directory = path.join(results_directory, "guardians")
    ballots_directory = path.join(results_directory, "encrypted_ballots")
    spoiled_directory = path.join(results_directory, "spoiled_ballots")

    make_directory(results_directory)
    manifest.to_json_file(MANIFEST_FILE_NAME, results_directory)
    context.to_json_file(CONTEXT_FILE_NAME, results_directory)
    constants.to_json_file(CONSTANTS_FILE_NAME, results_directory)

    make_directory(devices_directory)
    for device in devices:
        device_name = DEVICE_PREFIX + str(device.device_id)
        device.to_json_file(device_name, devices_directory)

    make_directory(guardian_directory)
    if guardian_records is not None:
        for guardian_record in guardian_records:
            set_name = GUARDIAN_PREFIX + guardian_record.guardian_id
            guardian_record.to_json_file(set_name, guardian_directory)

    make_directory(ballots_directory)
    for ballot in ciphertext_ballots:
        name = BALLOT_PREFIX + ballot.object_id
        ballot.to_json_file(name, ballots_directory)

    make_directory(spoiled_directory)
    for spoiled_ballot in spoiled_ballots:
        name = BALLOT_PREFIX + spoiled_ballot.object_id
        spoiled_ballot.to_json_file(name, spoiled_directory)

    ciphertext_tally.to_json_file(ENCRYPTED_TALLY_FILE_NAME, results_directory)
    plaintext_tally.to_json_file(TALLY_FILE_NAME, results_directory)


def publish_private_data(
    plaintext_ballots: Iterable[PlaintextBallot],
    ciphertext_ballots: Iterable[CiphertextBallot],
    guardian_records: Iterable[PrivateGuardianRecord],
    results_directory: str = RESULTS_DIR,
) -> None:
    """
    Publish the private data for an election.
    Useful for generating sample data sets.
    Do not use this in a production application.
    """

    private_directory = path.join(results_directory, "private")
    plaintext_ballots_directory = path.join(private_directory, "plaintext")
    encrypted_ballots_directory = path.join(private_directory, "encrypted")

    make_directory(results_directory)
    make_directory(private_directory)

    for guardian_record in guardian_records:
        guardian_name = GUARDIAN_PREFIX + guardian_record.guardian_id
        guardian_record.to_json_file(
            guardian_name, private_directory, strip_privates=False
        )

    make_directory(plaintext_ballots_directory)
    for plaintext_ballot in plaintext_ballots:
        name = PLAINTEXT_BALLOT_PREFIX + plaintext_ballot.object_id
        plaintext_ballot.to_json_file(name, plaintext_ballots_directory)

    make_directory(encrypted_ballots_directory)
    for ciphertext_ballot in ciphertext_ballots:
        name = BALLOT_PREFIX + ciphertext_ballot.object_id
        ciphertext_ballot.to_json_file(name, encrypted_ballots_directory)



# pylint: disable=import-outside-toplevel

from dataclasses import dataclass
from datetime import datetime
import re
from os import path
from typing import Any, cast, Type, TypeVar

from jsons import (
    dump,
    dumps,
    NoneType,
    load,
    loads,
    JsonsError,
    set_deserializer,
    set_serializer,
    set_validator,
    suppress_warnings,
    default_nonetype_deserializer,
)

S = TypeVar("S", bound="Serializable")
_T = TypeVar("_T")

JSON_FILE_EXTENSION: str = ".json"
WRITE: str = "w"
READ: str = "r"
JSON_PARSE_ERROR = '{"error": "Object could not be parsed due to json issue"}'
# TODO Issue #??: Jsons library incorrectly dumps class method
KEYS_TO_REMOVE = ["from_json", "from_json_file", "from_json_object", "_is_protocol_"]


@dataclass
class Serializable:
    """
    Serializable class with methods to convert to json
    """

    def to_json(self, strip_privates: bool = True) -> str:
        """
        Serialize to json string
        :param strip_privates: strip private variables
        :return: the json string representation of this object
        """
        return write_json(self, strip_privates)

    def to_json_object(self, strip_privates: bool = True) -> Any:
        """
        Serialize to json object
        :param strip_privates: strip private variables
        :return: the json representation of this object
        """
        return write_json_object(self, strip_privates)

    def to_json_file(
        self, file_name: str, file_path: str = "", strip_privates: bool = True
    ) -> None:
        """
        Serialize an object to a json file
        :param file_name: File name
        :param file_path: File path
        :param strip_privates: Strip private variables
        """
        write_json_file(self, file_name, file_path, strip_privates)

    @classmethod
    def from_json(cls: Type[S], data: str) -> S:
        """
        Deserialize the provided data string into the specified instance
        :param data: JSON string
        """
        return read_json(data, cls)

    @classmethod
    def from_json_object(cls: Type[S], data: object) -> S:
        """
        Deserialize the provided data object into the specified instance
        :param data: JSON object
        """
        return read_json_object(data, cls)

    @classmethod
    def from_json_file(cls: Type[S], file_name: str, file_path: str = "") -> S:
        """
        Deserialize the provided file into the specified instance
        :param file_name: File name
        :param file_path: File path
        """
        return read_json_file(cls, file_name, file_path)


def _remove_key(obj: Any, key_to_remove: str) -> Any:
    """
    Remove key from object recursively
    :param obj: Any object
    :param key_to_remove: key to remove
    """
    if isinstance(obj, dict):
        for key in list(obj.keys()):
            if key == key_to_remove:
                del obj[key]
            else:
                _remove_key(obj[key], key_to_remove)
    elif isinstance(obj, list):
        for i in reversed(range(len(obj))):
            if obj[i] == key_to_remove:
                del obj[i]
            else:
                _remove_key(obj[i], key_to_remove)


def write_json(object_to_write: object, strip_privates: bool = True) -> str:
    """
    Serialize to json string
    :param object_to_write: object to write to json
    :param strip_privates: strip private variables
    :return: the json string representation of this object
    """
    set_serializers()
    suppress_warnings()
    try:
        json_object = write_json_object(object_to_write, strip_privates)
        json_string = cast(
            str,
            dumps(
                json_object,
                strip_privates=strip_privates,
                strip_nulls=True,
                ensure_ascii=False,
            ),
        )
        return json_string
    except JsonsError:
        return JSON_PARSE_ERROR


def write_json_object(object_to_write: object, strip_privates: bool = True) -> object:
    """
    Serialize to json object
    :param object_to_write: object to write to json
    :param strip_privates: strip private variables
    :return: the json representation of this object
    """
    set_serializers()
    suppress_warnings()
    try:
        json_object = dump(
            object_to_write,
            strip_privates=strip_privates,
            strip_nulls=True,
            ensure_ascii=False,
        )
        for key in KEYS_TO_REMOVE:
            _remove_key(json_object, key)
        return json_object
    except JsonsError:
        return JSON_PARSE_ERROR


def write_json_file(
    object_to_write: object,
    file_name: str,
    file_path: str = "",
    strip_privates: bool = True,
) -> None:
    """
    Serialize json data string to json file
    :param object_to_write: object to write to json
    :param file_name: File name
    :param file_path: File path
    :param strip_privates: strip private variables
    """
    json_file_path: str = path.join(file_path, file_name + JSON_FILE_EXTENSION)
    with open(json_file_path, WRITE) as json_file:
        json_file.write(write_json(object_to_write, strip_privates))


def read_json(data: Any, class_out: Type[_T]) -> _T:
    """
    Deserialize json file to object
    :param data: Json file data
    :param class_out: Object type
    :return: Deserialized object
    """
    set_deserializers()
    return cast(_T, loads(data, class_out))


def read_json_object(data: Any, class_out: Type[_T]) -> _T:
    """
    Deserialize json file to object
    :param data: Json file data
    :param class_out: Object type
    :return: Deserialized object
    """
    set_deserializers()
    return cast(_T, load(data, class_out))


def read_json_file(class_out: Type[_T], file_name: str, file_path: str = "") -> _T:
    """
    Deserialize json file to object
    :param class_out: Object type
    :param file_name: File name
    :param file_path: File path
    :return: Deserialized object
    """
    set_deserializers()
    json_file_path: str = path.join(file_path, file_name + JSON_FILE_EXTENSION)
    with open(json_file_path, READ) as json_file:
        data = json_file.read()
        target: _T = read_json(data, class_out)
    return target


def set_serializers() -> None:
    """Set serializers for jsons to use to cast specific classes"""

    # Local import to minimize jsons usage across files
    from .group import ElementModP, ElementModQ

    set_serializer(lambda p, **_: str(p), ElementModP)
    set_serializer(lambda q, **_: str(q), ElementModQ)
    set_serializer(lambda dt, **_: dt.isoformat(), datetime)


def set_deserializers() -> None:
    """Set deserializers and validators for json to use to cast specific classes"""

    # Local import to minimize jsons usage across files
    from .group import ElementModP, ElementModQ, int_to_p_unchecked, int_to_q_unchecked

    set_deserializer(
        lambda p_as_int, cls, **_: int_to_p_unchecked(p_as_int), ElementModP
    )
    set_validator(lambda p: p.is_in_bounds(), ElementModP)

    set_deserializer(
        lambda q_as_int, cls, **_: int_to_q_unchecked(q_as_int), ElementModQ
    )
    set_validator(lambda q: q.is_in_bounds(), ElementModQ)

    set_deserializer(
        lambda none, cls, **_: None
        if none == "None"
        else default_nonetype_deserializer(none),
        NoneType,
    )

    set_deserializer(lambda dt, cls, **_: _deserialize_datetime(dt), datetime)


def _deserialize_datetime(value: str) -> datetime:
    """
    The `fromisoformat` function doesn't recognize the Z (Zulu) suffix
    to indicate UTC.  For compatibility with more external clients, we
    should allow it.
    """
    tz_corrected = re.sub("Z$", "+00:00", value)
    return datetime.fromisoformat(tz_corrected)



from collections.abc import Mapping

from typing import (
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
)


_T = TypeVar("_T")
_U = TypeVar("_U")


class DataStore(Generic[_T, _U]):
    """
    A lightweight convenience wrapper around a dictionary for data storage.
    This implementation defines the common interface used to access stored
    state elements.
    """

    _store: Dict[_T, _U]

    def __init__(self) -> None:
        self._store = {}

    def __iter__(self) -> Iterator:
        return iter(self._store.items())

    def all(self) -> List[_U]:
        """
        Get all `SubmittedBallot` from the store
        """
        return list(self._store.values())

    def clear(self) -> None:
        """
        Clear data from store
        """
        self._store.clear()

    def get(self, key: _T) -> Optional[_U]:
        """
        Get value in store
        :param key: key
        :return: value if found
        """
        return self._store.get(key)

    def items(self) -> Iterable[Tuple[_T, _U]]:
        """
        Gets all items in store as list
        :return: List of (key, value)
        """
        return self._store.items()

    def keys(self) -> Iterable[_T]:
        """
        Gets all keys in store as list
        :return: List of keys
        """
        return self._store.keys()

    def __len__(self) -> int:
        """
        Get length or count of store
        :return: Count in store
        """
        return len(self._store)

    def pop(self, key: _T) -> Optional[_U]:
        """
        Pop an object from the store if it exists.
        :param key: key
        """
        if key in self._store:
            return self._store.pop(key)
        return None

    def set(self, key: _T, value: _U) -> None:
        """
        Create or update a new value in store
        :param key: key
        :param value: value
        """
        self._store[key] = value

    def values(self) -> Iterable[_U]:
        """
        Gets all values in store as list
        :return: List of values
        """
        return self._store.values()


class ReadOnlyDataStore(Generic[_T, _U], Mapping):
    """
    A readonly view to a Data store
    """

    def __init__(self, data: DataStore[_T, _U]):
        self._data: DataStore[_T, _U] = data

    def __getitem__(self, key: _T) -> Optional[_U]:
        return self._data.get(key)

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator:
        return iter(self._data.items())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ReadOnlyDataStore):
            return False
        return ReadOnlyDataStore.__eq__(self, other)
